;;; dijkstra-kanten.scm
(load "breakpoint.scm") ; Ausgaben und bkpt

; Das Programm benoetigt die ausgegliederten Funktionen:
(load "prioritaetswarteschlange.scm")
; mit der Funktion (fuege-alle-ein <wegeliste> vor? <priows>)
; Das Praedikat vor? muss dafuer definiert werden:
(define
  (vor? weg-1 weg-2)
  (< (first weg-1) (first weg-2)))
; und
(load "gallenbacher-graph-kanten.scm")
; Verwendet wird davon hier die Liste "kanten"
; und
(load "wege.scm")
; Verwendet wird die Funktion (alle-nachfolge-wege <weg> <kanten>)
;;; Hinweis: Das Programm nutzt die Moeglichkeit, dass innere Funktionen
;;; auf die Parameter der aeusseren zugreifen koennen.

;;; =============== Entwurfsprinzip fuer Dijksta ===============
;;; Durch das Verwenden der Prioritaetswarteschlange ist gewaehrleistet,
;;; dass nur die kuerzesten Wege zu einem Knoten verwendet werden.
;;; Varianten:
; - Es wird nur der kuerzeste Weg zu einem Zielknoten gesucht.
; - Es werden alle kuerzesten Wege vom Startknoten aus gesucht.
;;; Entwurfentscheidung: Variante 1 wird bearbeitet.

;;; ===== Dijksta =====
; Loesung mit passender Aufrufhuelle
(define
  (dijkstra start-knoten ziel-knoten graph)
  (define
    (dijkstra-intern besucht kuerzeste-wege priows)
    ;(ausgabe (length graph) #\newline)
    (cond
      ((null? priows)
       #f)
      ((equal? (second (first priows)) ziel-knoten)
       (ausgabe (cons (first priows) kuerzeste-wege))
       (ausgabe "Loesung: ")
       (first priows))
      ((member (second (first priows)) besucht)
       (dijkstra-intern besucht kuerzeste-wege (rest priows)))
      (else
       (dijkstra-intern
        (aktualisiere-besuchtliste (second (first priows)) besucht)
        (aktualisisere-kuerzeste-wege besucht (first priows) kuerzeste-wege)
        (aktualisiere-priows priows graph)))
      )) ; Ende interne Funktion
  ; interner Aufruf:
  (dijkstra-intern
   (list start-knoten)     ; besucht
   '()                     ; kuerzeste-wege
   (fuege-alle-ein         ; Start-PrioWS
    (alle-nachfolge-wege
     (list 0 start-knoten) ; Start-'Weg'
     graph)
    vor?
    '())
   ))

;;; ===== Hilfsfunktion aktualisiere-besuchtliste =====
(define
  (aktualisiere-besuchtliste knoten besucht)
  (cond
   ((member knoten besucht)
    besucht)
   (else
   (cons knoten besucht))
   ))

;;; ===== Hilfsfunktion aktualisiere-priows =====
; Zum ersten Weg aus der Prioritaetswarteschlange werden alle weiter
; fuehrenden wege des Graphen bestimmt
; und in den Rest der priows eingefuegt
(define
  (aktualisiere-priows priows graph)
  (fuege-alle-ein
   (alle-nachfolge-wege
    (first priows)
    graph)
   vor?
   (rest priows)))

;;; ===== Hilfsfunktion aktualisisere-kuerzeste-wege =====
; Nur neu erreichte Knoten ergeben einen kuerzesten Weg
(define
  (aktualisisere-kuerzeste-wege besucht weg kuerzeste-wege)
  ;;(bkpt  besucht weg kuerzeste-wege)
  (cond
   ((member (second weg) besucht)
    kuerzeste-wege)
   (else
    (cons weg kuerzeste-wege))
   ))
;-----
; Tests dijkstra:
(dijkstra 'I 'O *kanten*) ; -> '(123 O F K P I)
;(dijkstra 'A 'E *kanten*) ; -> '(126 E K Y Z N A)
;(dijkstra 'P 'G *kanten*) ; -> '(75 G Y K P)
